﻿using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Windows.Forms;
using System.Text;
using System.Runtime.InteropServices;
using FXED.Classes.SyntaxObjects;

namespace FXED.CustomControls
{
    public class SyntaxEditor : RichTextBox
    {
        public const int WM_USER = 0x400;
        public const int WM_PAINT = 0xF;

        public const int EM_GETSCROLLPOS = (WM_USER + 221);
        public const int EM_SETSCROLLPOS = (WM_USER + 222);

        [DllImport("user32")]
        public static extern int SendMessage(IntPtr hwnd, int wMsg, int wParam, IntPtr lParam);
        [DllImport("user32")]
        public static extern int PostMessage(IntPtr hwnd, int wMsg, int wParam, int lParam);
        [DllImport("user32")]
        public static extern short GetKeyState(int nVirtKey);
        [DllImport("user32")]
        public static extern int LockWindowUpdate(IntPtr hwnd);

        private SeperaratorCollection mSeperators = new SeperaratorCollection();
        private HighLightDescriptorCollection intrinsic_descriptors = new HighLightDescriptorCollection();
        private HighLightDescriptorCollection object_descriptors = new HighLightDescriptorCollection();
        private HighLightDescriptorCollection parameter_descriptors = new HighLightDescriptorCollection();
        private bool mParsing = false;

        //Undo/Redo members
        private ArrayList mUndoList = new ArrayList();
        private Stack mRedoStack = new Stack();
        private bool mIsUndo = false;
        private UndoRedoInfo mLastInfo = new UndoRedoInfo("", new POINT(), 0);
        private int mMaxUndoRedoSteps = 50;

        public SeperaratorCollection Seperators
        {
            get
            {
                return mSeperators;
            }
        }

        public HighLightDescriptorCollection IntrinsicDescriptors
        {
            get { return intrinsic_descriptors; }
        }

        public HighLightDescriptorCollection ParameterDescriptors
        {
            get { return parameter_descriptors; }
        }

        public HighLightDescriptorCollection ObjectDescriptors
        {
            get { return object_descriptors; }
        }

        public SyntaxEditor() // Constructor
        {
            Seperators.Add(' ');
            Seperators.Add('\r');
            Seperators.Add('\n');
            Seperators.Add('\t');
            Seperators.Add(',');
            Seperators.Add('.');
            Seperators.Add('-');
            Seperators.Add('+');
            //Seperators.Add('*');
            //Seperators.Add('/');
            Seperators.Add('%');
            Seperators.Add('|');
            Seperators.Add('&');
            Seperators.Add('~');
            Seperators.Add('(');
            Seperators.Add(')');
            Seperators.Add('{');
            Seperators.Add('}');
            Seperators.Add('[');
            Seperators.Add(']');
            Seperators.Add('<');
            Seperators.Add('>');
            Seperators.Add(':');
            Seperators.Add(';');
            Seperators.Add('!');
            Seperators.Add('=');

            IntrinsicDescriptors.Add(new HighlightDescriptor("#", null, Properties.Settings.Default.PreProcessorColor, null, DescriptorType.ToEOL, DescriptorRecognition.StartsWith, "PreProcessor"));
            IntrinsicDescriptors.Add(new HighlightDescriptor("//", null, Properties.Settings.Default.CommentColor, null, DescriptorType.ToEOL, DescriptorRecognition.StartsWith, "Comment"));
            IntrinsicDescriptors.Add(new HighlightDescriptor("/*", "*/", Properties.Settings.Default.CommentColor, null, DescriptorType.ToCloseToken, DescriptorRecognition.StartsWith, "Comment"));
            IntrinsicDescriptors.Add(new HighlightDescriptor("\"", "\"", Properties.Settings.Default.StringColor, null, DescriptorType.ToCloseToken, DescriptorRecognition.StartsWith, "String"));
            IntrinsicDescriptors.Add(new HighlightDescriptor("'", "'", Properties.Settings.Default.CharacterColor, null, DescriptorType.ToCloseToken, DescriptorRecognition.StartsWith, "Character"));

            // keywords

            for (int i = 0; i < Classes.Variables.reserved_words.Length; i++)
            {
                IntrinsicDescriptors.Add(new HighlightDescriptor(Classes.Variables.reserved_words[i], Properties.Settings.Default.KeywordColor, null, DescriptorType.Word, DescriptorRecognition.WholeWord, "Keyword"));
            }

            // functions

            for (int i = 0; i < Classes.Variables.intrinsic_functions.Length; i++)
            {
                IntrinsicDescriptors.Add(new HighlightDescriptor(Classes.Variables.intrinsic_functions[i], Properties.Settings.Default.FunctionColor, null, DescriptorType.Word, DescriptorRecognition.WholeWord,"Function"));
            }

            // semantics

            if (Classes.Variables.my_effect.UseFxFile)
            {
                for (int i = 0; i < Classes.Variables.predefined_semantics.Length; i++)
                {
                    IntrinsicDescriptors.Add(new HighlightDescriptor(Classes.Variables.predefined_semantics[i], Properties.Settings.Default.SemanticColor, null, DescriptorType.Word, DescriptorRecognition.WholeWord, "Semantic"));
                }
            }
        }

        public void AddObject(string text)
        {
            ObjectDescriptors.Add(new HighlightDescriptor(text, Properties.Settings.Default.ComponentColor, null, DescriptorType.Word, DescriptorRecognition.WholeWord, "Object"));
        }

        public void AddParameter(string text)
        {
            ParameterDescriptors.Add(new HighlightDescriptor(text, Properties.Settings.Default.ParameterColor, null, DescriptorType.Word, DescriptorRecognition.WholeWord, "Parameter"));
        }

        public override string Text 
        { 
            get { return base.Text; }
            set { base.Text = value;} 
        }

        // methods

        public void UpdateFont()
        {
            this.Font = Properties.Settings.Default.EditorFont;
        }

        public void UpdateColors()
        {
            foreach (HighlightDescriptor hd in intrinsic_descriptors)
            {
                switch (hd.Type)
                {
                    case "Comment": hd.Color = Properties.Settings.Default.CommentColor; break;
                    case "String": hd.Color = Properties.Settings.Default.StringColor; break;
                    case "Character": hd.Color = Properties.Settings.Default.CharacterColor; break;
                    case "Keyword": hd.Color = Properties.Settings.Default.KeywordColor; break;
                    case "Function": hd.Color = Properties.Settings.Default.FunctionColor; break;
                    case "Semantic": hd.Color = Properties.Settings.Default.SemanticColor; break;
                    case "PreProcessor": hd.Color = Properties.Settings.Default.PreProcessorColor; break;
                }
            }

            foreach (HighlightDescriptor hd in object_descriptors)
            {
                 hd.Color = Properties.Settings.Default.ComponentColor; break;
            }

            foreach (HighlightDescriptor hd in parameter_descriptors)
            {
                hd.Color = Properties.Settings.Default.ParameterColor; break;
            }

            PerformHighlighting();
        }

        public void PerformHighlighting()
        {
            if (mParsing) return;
            mParsing = true;
            LockWindowUpdate(Handle);

            if (!mIsUndo)
            {
                mRedoStack.Clear();
                mUndoList.Insert(0, mLastInfo);
                this.LimitUndo();
                mLastInfo = new UndoRedoInfo(Text, GetScrollPos(), SelectionStart);
            }

            //Save scroll bar an cursor position, changeing the RTF moves the cursor and scrollbars to top positin
            POINT scrollPos = GetScrollPos();
            int cursorLoc = SelectionStart;

            //Created with an estimate of how big the stringbuilder has to be...
            StringBuilder sb = new
                StringBuilder((int)(Text.Length * 1.5 + 150));

            //Adding RTF header
            sb.Append(@"{\rtf1\fbidis\ansi\ansicpg1255\deff0\deflang1037{\fonttbl{");

            // Aggregrate functions/objects

            HighLightDescriptorCollection aggregation = new HighLightDescriptorCollection();
            aggregation.AddRange(intrinsic_descriptors.Data);
            aggregation.AddRange(object_descriptors.Data);
            aggregation.AddRange(parameter_descriptors.Data);

            //Font table creation
            int fontCounter = 0;
            Hashtable fonts = new Hashtable();
            AddFontToTable(sb, Font, ref fontCounter, fonts);
            foreach (HighlightDescriptor hd in aggregation)
            {
                if ((hd.Font != null) && !fonts.ContainsKey(hd.Font.Name))
                {
                    AddFontToTable(sb, hd.Font, ref fontCounter, fonts);
                }
            }

            sb.Append("}\n");

            //ColorTable

            sb.Append(@"{\colortbl ;");
            Hashtable colors = new Hashtable();
            int colorCounter = 1;
            AddColorToTable(sb, ForeColor, ref colorCounter, colors);
            AddColorToTable(sb, BackColor, ref colorCounter, colors);

            foreach (HighlightDescriptor hd in aggregation)
            {
                if (!colors.ContainsKey(hd.Color))
                {
                    AddColorToTable(sb, hd.Color, ref colorCounter, colors);
                }
            }

            //Parsing text

            sb.Append("}\n").Append(@"\viewkind4\uc1\pard\ltrpar");
            SetDefaultSettings(sb, colors, fonts);

            char[] sperators = mSeperators.GetAsCharArray();

            //Replacing "\" to "\\" for RTF...
            string[] lines = Text.Replace("\\", "\\\\").Replace("{", "\\{").Replace("}", "\\}").Split('\n');
            for (int lineCounter = 0; lineCounter < lines.Length; lineCounter++)
            {
                if (lineCounter != 0)
                {
                    AddNewLine(sb);
                }
                string line = lines[lineCounter];
                string[] tokens = line.Split(sperators);
                if (tokens.Length == 0)
                {
                    sb.Append(line);
                    AddNewLine(sb);
                    continue;
                }

                int tokenCounter = 0;
                for (int i = 0; i < line.Length; )
                {
                    char curChar = line[i];
                    if (mSeperators.Contains(curChar))
                    {
                        sb.Append(curChar);
                        i++;
                    }
                    else
                    {

                        string curToken = tokens[tokenCounter++];
                        bool bAddToken = true;
                        foreach (HighlightDescriptor hd in aggregation)
                        {
                            string compareStr = hd.Token;
                            bool match = false;

                            //Check if the highlight descriptor matches the current toker according to the DescriptoRecognision property.
                            switch (hd.DescriptorRecognition)
                            {
                                case DescriptorRecognition.WholeWord:
                                    if (curToken == compareStr)
                                    {
                                        match = true;
                                    }
                                    break;
                                case DescriptorRecognition.StartsWith:
                                    if (curToken.StartsWith(compareStr))
                                    {
                                        match = true;
                                    }
                                    break;
                                case DescriptorRecognition.Contains:
                                    if (curToken.IndexOf(compareStr) != -1)
                                    {
                                        match = true;
                                    }
                                    break;
                            }
                            if (!match)
                            {
                                //If this token doesn't match chech the next one.
                                continue;
                            }

                            //printing this token will be handled by the inner code, don't apply default settings...
                            bAddToken = false;

                            //Set colors to current descriptor settings.
                            SetDescriptorSettings(sb, hd, colors, fonts);

                            //Print text affected by this descriptor.
                            switch (hd.DescriptorType)
                            {
                                case DescriptorType.Word:
                                    sb.Append(line.Substring(i, curToken.Length));
                                    SetDefaultSettings(sb, colors, fonts);
                                    i += curToken.Length;
                                    break;
                                case DescriptorType.ToEOL:
                                    sb.Append(line.Remove(0, i));
                                    i = line.Length;
                                    SetDefaultSettings(sb, colors, fonts);
                                    break;
                                case DescriptorType.ToCloseToken:
                                    {
                                        int closeStart = i + hd.Token.Length;
                                        while ((line.IndexOf(hd.CloseToken, closeStart) == -1) && (lineCounter < lines.Length))
                                        {
                                            sb.Append(line.Remove(0, i));
                                            lineCounter++;
                                            if (lineCounter < lines.Length)
                                            {
                                                AddNewLine(sb);
                                                line = lines[lineCounter];
                                                i = closeStart = 0;
                                            }
                                            else
                                            {
                                                i = closeStart = line.Length;
                                            }

                                        }
                                        int closeIndex = line.IndexOf(hd.CloseToken, closeStart);
                                        if (closeIndex != -1)
                                        {
                                            sb.Append(line.Substring(i, closeIndex + hd.CloseToken.Length - i));
                                            line = line.Remove(0, closeIndex + hd.CloseToken.Length);
                                            tokenCounter = 0;
                                            tokens = line.Split(sperators);
                                            SetDefaultSettings(sb, colors, fonts); i = 0;
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                        if (bAddToken)
                        {
                            //Print text with default settings...
                            sb.Append(line.Substring(i, curToken.Length));
                            i += curToken.Length;
                        }
                    }
                }
            }

            Rtf = sb.ToString();

            //Restore cursor and scrollbars location.
            SelectionStart = cursorLoc;

            mParsing = false;

            SetScrollPos(scrollPos);
            LockWindowUpdate((IntPtr)0);
            Invalidate();
        }

        protected override void OnVScroll(EventArgs e)
        {
            if (mParsing) return;
            base.OnVScroll(e);
        }

        protected override void WndProc(ref Message m)
        {
            switch (m.Msg)
            {
                case WM_PAINT:
                    {
                        if (mParsing)
                        {
                            return;
                        }
                        break;
                    }
            }
            base.WndProc(ref m);
        }

        private void SetDefaultSettings(StringBuilder sb, Hashtable colors, Hashtable fonts)
        {
            SetColor(sb, ForeColor, colors);
            SetFont(sb, Font, fonts);
            SetFontSize(sb, (int)Font.Size);
            EndTags(sb);
        }

        private void SetDescriptorSettings(StringBuilder sb, HighlightDescriptor hd, Hashtable colors, Hashtable fonts)
        {
            SetColor(sb, hd.Color, colors);
            if (hd.Font != null)
            {
                SetFont(sb, hd.Font, fonts);
                SetFontSize(sb, (int)hd.Font.Size);
            }
            EndTags(sb);
        }

        private void SetColor(StringBuilder sb, Color color, Hashtable colors)
        {
            sb.Append(@"\cf").Append(colors[color]);
        }
 
        private void SetBackColor(StringBuilder sb, Color color, Hashtable colors)
        {
            sb.Append(@"\cb").Append(colors[color]);
        }

        private void SetFont(StringBuilder sb, Font font, Hashtable fonts)
        {
            if (font == null) return;
            sb.Append(@"\f").Append(fonts[font.Name]);
        }

        private void SetFontSize(StringBuilder sb, int size)
        {
            sb.Append(@"\fs").Append(size * 2);
        }

        private void AddNewLine(StringBuilder sb)
        {
            sb.Append("\\par\n");
        }

        private void EndTags(StringBuilder sb)
        {
            sb.Append(' ');
        }

        private void AddFontToTable(StringBuilder sb, Font font, ref int counter, Hashtable fonts)
        {
            sb.Append(@"\f").Append(counter).Append(@"\fnil\fcharset0").Append(font.Name).Append(";}");
            fonts.Add(font.Name, counter++);
        }

        private void AddColorToTable(StringBuilder sb, Color color, ref int counter, Hashtable colors)
        {
            sb.Append(@"\red").Append(color.R).Append(@"\green").Append(color.G).Append(@"\blue")
                .Append(color.B).Append(";");
            colors.Add(color, counter++);
        }

        private unsafe POINT GetScrollPos()
        {
            POINT res = new POINT();
            IntPtr ptr = new IntPtr(&res);
            SendMessage(Handle, EM_GETSCROLLPOS, 0, ptr);
            return res;
        }

        private unsafe void SetScrollPos(POINT point)
        {
            IntPtr ptr = new IntPtr(&point);
            SendMessage(Handle, EM_SETSCROLLPOS, 0, ptr);
        }

        public new bool CanUndo
        {
            get
            {
                return mUndoList.Count > 0;
            }
        }
        public new bool CanRedo
        {
            get
            {
                return mRedoStack.Count > 0;
            }
        }

        private void LimitUndo()
        {
            while (mUndoList.Count > mMaxUndoRedoSteps)
            {
                mUndoList.RemoveAt(mMaxUndoRedoSteps);
            }
        }

        public new void Undo()
        {
            if (!CanUndo) { return; }
                
            mIsUndo = true;
            mRedoStack.Push(new UndoRedoInfo(Text, GetScrollPos(), SelectionStart));
            UndoRedoInfo info = (UndoRedoInfo)mUndoList[0];
            mUndoList.RemoveAt(0);
            Text = info.Text;
            SelectionStart = info.CursorLocation;
            SetScrollPos(info.ScrollPos);
            mLastInfo = info;
            mIsUndo = false;
        }

        public new void Redo()
        {
            if (!CanRedo) { return; }
                
            mIsUndo = true;
            mUndoList.Insert(0, new UndoRedoInfo(Text, GetScrollPos(), SelectionStart));
            LimitUndo();
            UndoRedoInfo info = (UndoRedoInfo)mRedoStack.Pop();
            Text = info.Text;
            SelectionStart = info.CursorLocation;
            SetScrollPos(info.ScrollPos);
            mIsUndo = false;
        }
    }
}
